/*
 *	BOARD.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 *
 *  Traduccion al castellano por Jose Carlos Martinez Galan
 *	por sugerencia de Roger Davis
 *
 */


#include "defs.h"
#include "data.h"
#include "protos.h"


/* init() establece la posicion inicial del tablero */

void init()
{
	int i;

	for (i = 0; i < 64; ++i) {
		color[i] = init_color[i];
		piece[i] = init_piece[i];
	}
	side = LIGHT;	// Le toca jugar a las blancas
	xside = DARK;
	castle = 15;	// Continene la informacion sobre los enroques. Ambos bandos pueden enrocar en ambos lados
					// cada uno de los 4 bits derechos representa la posibilidad o no de enrocar
	ep = -1;		// Al comenzar, no hay casilla de "al paso"
	fifty = 0;		// Para llevar la cuenta de la regla de las 50 jugadas
	ply = 0;
	gen_begin[0] = 0; // Inicio de la lista de jugadas generadas
	undos = 0;		// jugadas retrocedidas
}


/* in_check() (en jaque) devuelve TRUE si el bando s est en jaque
   y FALSE en caso contrario. Simplemente recorre todo el tablero
   para buscar el rey del bando s y llama a la funcion attack
   para ver si esta siendo atacado */

BOOL in_check(int s)
{
	int i;

	for (i = 0; i < 64; ++i)
		if (color[i] == s && piece[i] == KING)
			return attack(i, s ^ 1);
	return TRUE;  /* no deberia llegar aqui nunca */
}


/* attack() devuelve TRUE si la casilla sq esta atacada por el
   bando s, y FALSE en caso contrario */

BOOL attack(int sq, int s)
{
	int i, j, n;

	for (i = 0; i < 64; ++i)
		if (color[i] == s) {
			if (piece[i] == PAWN) {
				if (s == LIGHT) {
					if (GetFile(i) != 0 && i - 9 == sq)
						return TRUE;
					if (GetFile(i) != 7 && i - 7 == sq)
						return TRUE;
				}
				else {
					if (GetFile(i) != 0 && i + 7 == sq)
						return TRUE;
					if (GetFile(i) != 7 && i + 9 == sq)
						return TRUE;
				}
			}
			else
				for (j = 0; j < offsets[piece[i]]; ++j)
					for (n = i;;) {
						n = mailbox[mailbox64[n] + offset[piece[i]][j]];
						if (n == -1)
							break;
						if (n == sq)
							return TRUE;
						if (color[n] != EMPTY)
							break;
						if (!slide[piece[i]])
							break;
					}
		}
	return FALSE;
}


/* gen() genera movimientos pseudo-legales para la posicion actual.
   Recorre el tablero para encontrar piezas propias y determina
   que casillas atacan. Cuando encuentra una combinacion de
   pieza/casilla, llama a gen_push para almacenar el movimiento
   en la "pila de movimientos" */

void gen()
{
	int i, j, n;

	/* Hasta ahora, no tenemos jugadas para el ply actual */
	gen_end[ply] = gen_begin[ply];
	// gen_begin marca el inicio de las jugadas legales disponibles en el presente ply
	// gen_end marca el final del mismo

	for (i = 0; i < 64; ++i)
		if (color[i] == side) {
			if (piece[i] == PAWN) { // Los peones tiene una generacion manual de las jugadas
				if (side == LIGHT) {
					if (GetFile(i) != 0 && color[i - 9] == DARK)
						gen_push(i, i - 9, 17);
					if (GetFile(i) != 7 && color[i - 7] == DARK)
						gen_push(i, i - 7, 17);
					if (color[i - 8] == EMPTY) {
						gen_push(i, i - 8, 16);
						if (i >= 48 && color[i - 16] == EMPTY)
							gen_push(i, i - 16, 24);
					}
				}
				else {
					if (GetFile(i) != 0 && color[i + 7] == LIGHT)
						gen_push(i, i + 7, 17);
					if (GetFile(i) != 7 && color[i + 9] == LIGHT)
						gen_push(i, i + 9, 17);
					if (color[i + 8] == EMPTY) {
						gen_push(i, i + 8, 16);
						if (i <= 15 && color[i + 16] == EMPTY)
							gen_push(i, i + 16, 24);
					}
				}
			}
			else // El resto de las piezas se generan mediante la tecnica "mailbox"
				for (j = 0; j < offsets[piece[i]]; ++j)
					for (n = i;;) {
						n = mailbox[mailbox64[n] + offset[piece[i]][j]];
						if (n == -1)
							break;
						if (color[n] != EMPTY) {
							if (color[n] == xside)
								gen_push(i, n, 1);
							break;
						}
						gen_push(i, n, 0);
						if (!slide[piece[i]])
							break;
					}
		}

	/* generar enroques */
	if (side == LIGHT) {
		if (castle & 1)
			gen_push(E1, G1, 2);
		if (castle & 2)
			gen_push(E1, C1, 2);
	}
	else {
		if (castle & 4)
			gen_push(E8, G8, 2);
		if (castle & 8)
			gen_push(E8, C8, 2);
	}
	
	/* generar capturas al paso */
	if (ep != -1) {
		if (side == LIGHT) {
			if (GetFile(ep) != 0 && color[ep + 7] == LIGHT && piece[ep + 7] == PAWN)
				gen_push(ep + 7, ep, 21);
			if (GetFile(ep) != 7 && color[ep + 9] == LIGHT && piece[ep + 9] == PAWN)
				gen_push(ep + 9, ep, 21);
		}
		else {
			if (GetFile(ep) != 0 && color[ep - 9] == DARK && piece[ep - 9] == PAWN)
				gen_push(ep - 9, ep, 21);
			if (GetFile(ep) != 7 && color[ep - 7] == DARK && piece[ep - 7] == PAWN)
				gen_push(ep - 7, ep, 21);
		}
	}

	/* las jugadas del proximo ply deben comenzar donde acaban las de este */
	gen_begin[ply + 1] = gen_end[ply];
}


/* gen_caps() (generar capturas) es basicamente una copia de gen()
   modificada para generar solo capturas y promociones. Se usa en
   la quiescence search */

void gen_caps()
{
	int i, j, n;

	gen_end[ply] = gen_begin[ply];
	for (i = 0; i < 64; ++i)
		if (color[i] == side) {
			if (piece[i]==PAWN) {
				if (side == LIGHT) {
					if (GetFile(i) != 0 && color[i - 9] == DARK)
						gen_push(i, i - 9, 17);
					if (GetFile(i) != 7 && color[i - 7] == DARK)
						gen_push(i, i - 7, 17);
					if (i <= 15 && color[i - 8] == EMPTY)
						gen_push(i, i - 8, 16);
				}
				if (side == DARK) {
					if (GetFile(i) != 0 && color[i + 7] == LIGHT)
						gen_push(i, i + 7, 17);
					if (GetFile(i) != 7 && color[i + 9] == LIGHT)
						gen_push(i, i + 9, 17);
					if (i >= 48 && color[i + 8] == EMPTY)
						gen_push(i, i + 8, 16);
				}
			}
			else
				for (j = 0; j < offsets[piece[i]]; ++j)
					for (n = i;;) {
						n = mailbox[mailbox64[n] + offset[piece[i]][j]];
						if (n == -1)
							break;
						if (color[n] != EMPTY) {
							if (color[n] == xside)
								gen_push(i, n, 1);
							break;
						}
						if (!slide[piece[i]])
							break;
					}
		}
	if (ep != -1) {
		if (side == LIGHT) {
			if (GetFile(ep) != 0 && color[ep + 7] == LIGHT && piece[ep + 7] == PAWN)
				gen_push(ep + 7, ep, 21);
			if (GetFile(ep) != 7 && color[ep + 9] == LIGHT && piece[ep + 9] == PAWN)
				gen_push(ep + 9, ep, 21);
		}
		else {
			if (GetFile(ep) != 0 && color[ep - 9] == DARK && piece[ep - 9] == PAWN)
				gen_push(ep - 9, ep, 21);
			if (GetFile(ep) != 7 && color[ep - 7] == DARK && piece[ep - 7] == PAWN)
				gen_push(ep - 7, ep, 21);
		}
	}
	gen_begin[ply + 1] = gen_end[ply];
}


/* gen_push() incluye una jugada en la pila de movimientos,
   a menos que sea una promocion de peon, que debe ser
   manejada por gen_promote(). Tambien asigna un valor a la
   jugada para la ordenacion de jugadas en la busqueda alfa-beta.
   Si el movimiento es una captura, utiliza MVV/LVA (Victima
   Mas Valiosa/Atacante Menos Valioso). En caso contrario,
   emplea la heuristica de historia (cada jugada aumenta su
   valor cada vez que, en algun lugar del arbol, se comprueba
   que es buena). Notese que las capturas suman 1,000,000 a su
   valor para asegurar que siempre quedan por encima de las
   demas jugadas */

void gen_push(int from, int to, int bits)
{
	gen_t *g;
	
	if (bits & 16) {
		if (side == LIGHT) {
			if (to <= H8) {
				gen_promote(from, to, bits);
				return;
			}
		}
		else {
			if (to >= A1) {
				gen_promote(from, to, bits);
				return;
			}
		}
	}
	g = &gen_dat[gen_end[ply]];
	++gen_end[ply];
	g->m.b.from = (char)from;
	g->m.b.to = (char)to;
	g->m.b.promote = 0;
	g->m.b.bits = (char)bits;
	if (color[to] != EMPTY)
		g->score = 1000000 + (piece[to] * 10) - piece[from];
	else
		g->score = history[from][to];
}


/* gen_promote() es como gen_push(), pero incluye 4 movimientos
   en la pila de jugadas, una por cada posible pieza de promocion */

void gen_promote(int from, int to, int bits)
{
	int i;
	gen_t *g;
	
	for (i = KNIGHT; i <= QUEEN; ++i) {
		g = &gen_dat[gen_end[ply]];
		++gen_end[ply];
		g->m.b.from = (char)from;
		g->m.b.to = (char)to;
		g->m.b.promote = (char)i;
		g->m.b.bits = (char)(bits | 32);
		g->score = 1000000 + (i * 10);
	}
}


/* makemove() realiza un movimiento. Si el movimiento es
   ilegal, lo deshace y devuelve FALSE. En caso contrario,
   devuelve TRUE */

BOOL makemove(move_bytes m)
{
	
	/* comprobar si es legal enrocar, y mover, en ese caso,
	   la torre (el rey se mueve en el codigo normal del
	   movimiento */
	if (m.bits & 2) { // Este bit indica que el movimiento es un enroque
		int from, to;

		if (in_check(side)) // No se puede enrocar cuando estamos en jaque
			return FALSE;
		switch (m.to) {
			case 62:
				if (color[F1] != EMPTY || color[G1] != EMPTY ||
						attack(F1, xside) || attack(G1, xside))
					return FALSE;
				from = H1;
				to = F1;
				break;
			case 58:
				if (color[B1] != EMPTY || color[C1] != EMPTY || color[D1] != EMPTY ||
						attack(C1, xside) || attack(D1, xside))
					return FALSE;
				from = A1;
				to = D1;
				break;
			case 6:
				if (color[F8] != EMPTY || color[G8] != EMPTY ||
						attack(F8, xside) || attack(G8, xside))
					return FALSE;
				from = H8;
				to = F8;
				break;
			case 2:
				if (color[B8] != EMPTY || color[C8] != EMPTY || color[D8] != EMPTY ||
						attack(C8, xside) || attack(D8, xside))
					return FALSE;
				from = A8;
				to = D8;
				break;
			default:  /* nunca debe llegar aqui */
				from = -1;
				to = -1;
				break;
		}
		color[to] = color[from];
		piece[to] = piece[from];
		color[from] = EMPTY;
		piece[from] = EMPTY;
	}
	// Si era un enroque, lo que queda mueve el rey. En caso contrario,
	// es un movimiento normal de pieza

	/* Almacenar informacion de retroceso, para poder volver atras despues */
	hist_dat[ply].m.b = m;
	hist_dat[ply].capture = piece[(int)m.to];
	hist_dat[ply].castle = castle;
	hist_dat[ply].ep = ep;
	hist_dat[ply].fifty = fifty;
	++ply;

	/* Actualizar las variables de enroque, al paso
	   y regla de las 50 jugadas */
	castle &= castle_mask[(int)m.from] & castle_mask[(int)m.to];
	if (m.bits & 8) {
		if (side == LIGHT)
			ep = m.to + 8;
		else
			ep = m.to - 8;
	}
	else
		ep = -1;
	if (m.bits & 17)
		fifty = 0;
	else
		++fifty;

	/* mover la pieza */
	color[(int)m.to] = side;
	if (m.bits & 32)
		piece[(int)m.to] = m.promote;
	else
		piece[(int)m.to] = piece[(int)m.from];
	color[(int)m.from] = EMPTY;
	piece[(int)m.from] = EMPTY;

	/* eliminar el peon sobrante, si era una captura al paso */
	if (m.bits & 4) {
		if (side == LIGHT) {
			color[m.to + 8] = EMPTY;
			piece[m.to + 8] = EMPTY;
		}
		else {
			color[m.to - 8] = EMPTY;
			piece[m.to - 8] = EMPTY;
		}
	}

	/* Cambiar lados y comprobar la legalidad de la posicion
	   resultante (si podemos capturar el rey contrario, la
	   jugada es ilegal y debemos retrocederla) */
	side ^= 1;
	xside ^= 1;
	if (in_check(xside)) {
		takeback();
		return FALSE;
	}
	return TRUE;
}


/* takeback() es muy similar a makemove(), solo que al reves :) */

void takeback()
{
	move_bytes m;

	side ^= 1;
	xside ^= 1;
	--ply;
	m = hist_dat[ply].m.b;
	castle = hist_dat[ply].castle;
	ep = hist_dat[ply].ep;
	fifty = hist_dat[ply].fifty;
	color[(int)m.from] = side;
	if (m.bits & 32)
		piece[(int)m.from] = PAWN;
	else
		piece[(int)m.from] = piece[(int)m.to];
	if (hist_dat[ply].capture == EMPTY) {
		color[(int)m.to] = EMPTY;
		piece[(int)m.to] = EMPTY;
	}
	else {
		color[(int)m.to] = xside;
		piece[(int)m.to] = hist_dat[ply].capture;
	}
	if (m.bits & 2) {
		int from, to;

		switch(m.to) {
			case 62:
				from = F1;
				to = H1;
				break;
			case 58:
				from = D1;
				to = A1;
				break;
			case 6:
				from = F8;
				to = H8;
				break;
			case 2:
				from = D8;
				to = A8;
				break;
			default:  /* no debe llegar aqui */
				from = -1;
				to = -1;
				break;
		}
		color[to] = side;
		piece[to] = ROOK;
		color[from] = EMPTY;
		piece[from] = EMPTY;
	}
	if (m.bits & 4) {
		if (side == LIGHT) {
			color[m.to + 8] = xside;
			piece[m.to + 8] = PAWN;
		}
		else {
			color[m.to - 8] = xside;
			piece[m.to - 8] = PAWN;
		}
	}
}
